import { YOUTUBE_CLIENT_ID, YOUTUBE_CLIENT_SECRET, YOUTUBE_PLAYLIST_ID } from '$env/static/private';
import { Logger } from '$lib/log';
import type { OauthResponse } from '$lib/mastodon/response';
import type { SongInfo } from '$lib/odesliResponse';
import { OauthPlaylistAdder } from './oauthPlaylistAdder';
import type { PlaylistAdder } from './playlistAdder';

export class YoutubePlaylistAdder extends OauthPlaylistAdder implements PlaylistAdder {
  public constructor() {
    super('https://www.googleapis.com/youtube/v3', 'yt_auth_token');
    this.logger = new Logger('YoutubePlaylistAdder');
  }

  public constructAuthUrl(redirectUri: URL): URL {
    this.redirectUri = redirectUri;
    let additionalParameters = new Map([
      ['access_type', 'offline'],
      ['include_granted_scopes', 'false']
    ]);
    const endpoint = 'https://accounts.google.com/o/oauth2/v2/auth';
    return this.constructAuthUrlInternal(
      endpoint,
      YOUTUBE_CLIENT_ID,
      'https://www.googleapis.com/auth/youtube',
      redirectUri,
      additionalParameters
    );
  }

  public async receivedAuthCode(code: string, url: URL) {
    this.logger.debug('received code');
    this.redirectUri = url;
    const tokenUrl = new URL('https://oauth2.googleapis.com/token');
    await this.receivedAuthCodeInternal(
      tokenUrl,
      YOUTUBE_CLIENT_ID,
      code,
      url,
      YOUTUBE_CLIENT_SECRET
    );
    const token = await this.refreshToken();
    if (token == null) {
      return;
    }
    try {
      this.logger.debug('Checking authorized token');
      const res = await fetch(this.apiBase + '/channels?part=id&mine=true', {
        method: 'GET',
        headers: { Authorization: `${token.token_type} ${token.access_token}` }
      }).then((r) => r.json());
      this.logger.debug('Checked authorized token', res);
    } catch (e) {
      this.logger.debug('Error checking authorized token', e);
    }
  }

  private async refreshToken(force: boolean = false): Promise<OauthResponse | null> {
    const tokenInfo = await this.shouldRefreshToken();
    if (tokenInfo == null) {
      return null;
    }
    let token = tokenInfo.token;
    if (!tokenInfo.refresh && !force) {
      return token;
    }
    if (!token.refresh_token) {
      this.logger.error('Need to refresh access token, but no refresh token provided');
      return null;
    }

    const tokenUrl = new URL('https://oauth2.googleapis.com/token');
    return await this.requestRefreshToken(
      tokenUrl,
      YOUTUBE_CLIENT_ID,
      token.refresh_token,
      this.getRedirectUri('ytauth').toString(),
      YOUTUBE_CLIENT_SECRET
    );
  }

  public async addToPlaylist(song: SongInfo) {
    await this.addToPlaylistRetry(song);
  }

  private async addToPlaylistRetry(song: SongInfo, remaning: number = 3) {
    if (remaning < 0) {
      this.logger.error('max retries reached, song will not be added to spotify playlist');
    }
    this.logger.debug('addToYoutubePlaylist');
    const token = await this.refreshToken();
    if (token == null) {
      return;
    }

    if (!YOUTUBE_PLAYLIST_ID || YOUTUBE_PLAYLIST_ID === 'CHANGE_ME') {
      this.logger.debug('no playlist ID configured');
      return;
    }
    if (!song.youtubeUrl) {
      this.logger.info('Skip adding song to YT playlist, no youtube Url', song);
      return;
    }

    const songUrl = new URL(song.youtubeUrl);
    const youtubeId = songUrl.searchParams.get('v');
    if (!youtubeId) {
      this.logger.debug(
        'Skip adding song to YT playlist, could not extract YT id from URL',
        song.youtubeUrl
      );
      return;
    }
    this.logger.debug('Found YT id from URL', song.youtubeUrl, youtubeId);

    const playlistItemsUrl = new URL(this.apiBase + '/playlistItems');
    playlistItemsUrl.searchParams.append('videoId', youtubeId);
    playlistItemsUrl.searchParams.append('playlistId', YOUTUBE_PLAYLIST_ID);
    playlistItemsUrl.searchParams.append('part', 'id');
    const existingPlaylistItem = await fetch(playlistItemsUrl, {
      headers: { Authorization: `${token.token_type} ${token.access_token}` }
    }).then((r) => r.json());
    if (existingPlaylistItem.pageInfo && existingPlaylistItem.pageInfo.totalResults > 0) {
      this.logger.info('Item already in playlist', existingPlaylistItem);
      return;
    }

    const addItemUrl = new URL(this.apiBase + '/playlistItems');
    addItemUrl.searchParams.append('part', 'snippet');
    const options: RequestInit = {
      method: 'POST',
      headers: { Authorization: `${token.token_type} ${token.access_token}` },
      body: JSON.stringify({
        snippet: {
          playlistId: YOUTUBE_PLAYLIST_ID,
          resourceId: {
            videoId: youtubeId,
            kind: 'youtube#video'
          }
        }
      })
    };
    const resp = await fetch(addItemUrl, options);
    const respObj = await resp.json();
    this.logger.info('Added to playlist', youtubeId, song.title);
    if (respObj.error) {
      this.logger.error('Add to playlist failed', respObj.error.errors);
      if (respObj.error.errors && respObj.error.errors[0].reason === 'authError') {
        this.logger.info('Refreshing auth token');
        const token = await this.refreshToken(true);
        if (token == null) {
          return;
        }
        this.addToPlaylistRetry(song, remaning--);
      }
    }
  }
}
